package checkers.types;

import java.lang.annotation.Annotation;
import java.util.*;

import javax.lang.model.element.Element;
import javax.lang.model.type.*;

import com.sun.source.tree.Tree;

/**
 * AnnotatedClassType represents a type along with all its annotations.
 * Given a use of a type, clients can obtain an AnnotatedClassType and
 * query which annotations the type has.
 * <p>
 *
 * (This class abstracts the sources of the annotations: written
 * directly on the type, inherited from the supertype, or specified via a
 * "default" annotation.)
 */
public class AnnotatedClassType {

    /** The element node related to this annotated type. */
    protected Element element;

    /** The tree node related to this annotated type. */
    protected Tree tree;

    /** Maps annotation locations (raw type, generic/array types) to sets of
     * annotations present in those locations. */
    protected final Map<AnnotationLocation, Set<Class<? extends Annotation>>> annos;

    /** Annotation types treated as though they were written on the class type. */
    protected final Set<Class<? extends Annotation>> includes;

    /** Annotation types that should be ignored if present. */
    protected final Set<Class<? extends Annotation>> excludes;

    /**
     * Creates an annotated class type.
     */
    protected AnnotatedClassType() {

        // Initialize annotation storage.
        this.annos = new HashMap<AnnotationLocation, Set<Class<? extends Annotation>>>();
        this.includes = new HashSet<Class<? extends Annotation>>();
        this.excludes = new HashSet<Class<? extends Annotation>>();
    }

    /**
     * Determines whether the annotated type contains an annotation (either
     * directly or indirectly) with the given class (on the raw type only).
     *
     * @param annotation
     *            the class token for the annotation to check for
     * @return true if the given annotation is present on this annotated type,
     *         false otherwise
     * @deprecated use {@link AnnotatedClassType#hasAnnotationAt} instead
     */
    @Deprecated
    public boolean isAnnotatedWith(Class<? extends Annotation> annotation) {

        if (this.hasAnnotationAt(annotation, AnnotationLocation.RAW))
            return true;

        return false;
    }

    /**
     * Determines whether the annotated type has an inner (generic or array) 
     * type annotation with the given class at the given location.
     *
     * @param annotation
     *            the annotation to check for
     * @param location
     *            the location to check
     * @return true if the given annotation is present at the given location
     *         false otherwise
     */
    public boolean hasAnnotationAt(Class<? extends Annotation> annotation,
            AnnotationLocation location) {

        // If this type explicitly excludes this annotation type, regard it
        // as !isAnnotatedWith no matter what.
        if (location.equals(AnnotationLocation.RAW) 
                && excludes.contains(annotation))
            return false;

        if (location.equals(AnnotationLocation.RAW)
                && includes.contains(annotation))
            return true;

        if (annos.containsKey(location)
                && annos.get(location).contains(annotation))
            return true;

        return false;
    }

    /**
     * Determines whether the type backing this annotated type has a wildcard
     * ("?") at location specified by the given integer list.
     *
     * @param location
     *            location at which to check for a wildcard
     * @return true if there is a wildcard at the given location, false
     *         otherwise
     */
    public boolean hasWildcardAt(AnnotationLocation location) {

        if (element == null)
            return false;

        TypeMirror type = element.asType();

        if (!(type instanceof DeclaredType))
            return false;

        Queue<Integer> queue = new LinkedList<Integer>();
        queue.addAll(location.asList());

        // Walk down the generic type according to the integer list,
        // locating the parameter that the list specifies.

        TypeMirror current = type;
        while (!queue.isEmpty()) {
            if (current instanceof DeclaredType) {
                DeclaredType declared = (DeclaredType) current;
                int index = queue.poll();
                if (declared.getTypeArguments().size() <= index) {
                    //System.err.println("INVALID: " + declared);
                    return false;
                }
                current = declared.getTypeArguments().get(index);
            } else
                throw new RuntimeException("current: " + current.getKind());
        }

        // Check if the thing we've located is a wildcard.
        if (current instanceof WildcardType)
            return true;

        return false;
    }

    /**
     * Adds the given annotation, represented by an {@link InternalAnnotation},
     * to this class type at the location provided by the
     * {@link InternalAnnotation}.
     *
     * @param annotation the annotation to add
     */
    public void annotate(InternalAnnotation annotation) {
        annotateAt(annotation.getLocation(), annotation.getAnnotationClass());
    }

    /**
     * Manually adds an annotation to the raw type of this class type. It is
     * primarily useful for program elements that don't have symbols from which
     * annotations could be extracted (like typecasts or object creation
     * expressions).
     *
     * @param annotation
     *            the class token for the annotation type
     * @deprecated use {@link AnnotatedClassType#annotateAt} instead
     */
    @Deprecated
    public void annotate(Class<? extends Annotation> annotation) {

        if (annotation == null)
            throw new IllegalArgumentException("can't annotate type with null");

        this.annotateAt(AnnotationLocation.RAW, annotation);
    }

    /**
     * Manually adds an annotation to the inner (generic or array) type for this
     * class type at the specified location. It is primarily useful for program
     * elements that don't have symbols from which annotations could be 
     * extracted (like object creation expressions).
     *
     * @param location
     *            the location at which to add the annotation
     * @param annotation
     *            the class token for the annotation type
     */
    public void annotateAt(AnnotationLocation location,
            Class<? extends Annotation> annotation) {

        if (annotation == null)
            throw new IllegalArgumentException("can't annotate type with null");

        // Initialize that location with an empty set if nothing is there.
        if (!this.annos.containsKey(location))
            this.annos.put(location, new HashSet<Class<? extends Annotation>>());

        this.annos.get(location).add(annotation);
    }

    /**
     * Forces an annotation of the given type to be included among the
     * annotations checked by {@link AnnotatedClassType#isAnnotatedWith}. This is useful for
     * treating class types as having an annotation even though they don't
     * (e.g., a primitive type that is implicitly {@code @NonNull}.
     *
     * @param annotation
     *            the annotation to include among this type's annotations
     */
    public void include(Class<? extends Annotation> annotation) {

        if (annotation == null)
            throw new IllegalArgumentException("can't include null");

        this.includes.add(annotation);
    }

    /**
     * Forces an annotation of the given type to be skipped by
     * {@code isAnnotatedWith}. Useful in conjunction with annotation defaults
     * (e.g., if variables are {@code @NonNull} by default and the checker finds
     * a {@code @Nullable} variable, it can simply exclude {@code @NonNull}
     * rather than handle {@code @Nullable} as a special case).
     *
     * @param annotation
     *            the annotation to exclude from this type's annotations
     */
    public void exclude(Class<? extends Annotation> annotation) {

        if (annotation == null)
            throw new IllegalArgumentException("can't exclude null");

        this.excludes.add(annotation);
    }

    /**
     * Sets this type's element reference to the given element.
     *
     * @param e
     *            the element to associate with this type
     */
    void setElement(Element e) {
        this.element = e;
    }

    /**
     * @return the element associated with this type
     */
    public Element getElement() {
        return this.element;
    }

    /**
     * Sets this type's tree reference to the given tree.
     *
     * @param t the tree to associate with this type
     */
    void setTree(Tree t) {
        this.tree = t;
    }

    /**
     * @return the tree associated with this type
     */
    public Tree getTree() {
        return tree;
    }

    /**
     * @return the locations of all annotations in this class type
     *
     * @todo should take an annotation argument
     */
    public Set<AnnotationLocation> getAnnotatedLocations() {
        return Collections.unmodifiableSet(this.annos.keySet());
    }

    /**
     * @return the locations of all annotations in this class types, excluding
     * the location of annotations on raw types ({@link
     * AnnotationLocation#RAW}).
     */
    public Set<AnnotationLocation> getInnerLocations() {
        Set<AnnotationLocation> locs = this.getAnnotatedLocations();
        if (!locs.contains(AnnotationLocation.RAW))
            return locs;

        Set<AnnotationLocation> inner = new HashSet<AnnotationLocation>();
        inner.addAll(locs);
        inner.remove(AnnotationLocation.RAW);
        return Collections.unmodifiableSet(inner);
    }

    @Override
    public String toString() {
        return this.annos + " " + this.includes + " " + this.excludes;
    }
}
